import numpy as np

from bird import logger
from bird.utilities.mathtools import bissection


def make_walls_from_topo(topo_dict: dict) -> dict:
    """
    Find block cylindrical coordinates of all the wall blocks

    Parameters
    ----------
    topo_dict: dict
        Dictionary input that describes the topology

    Returns
    ----------
    wall_dict: dict
        Dictionary of wall coordinates
    """

    r_wall = []
    l_wall = []
    try:
        elements = topo_dict["Walls"]
        for element in elements:
            for block in elements[element]:
                r_wall.append(block["R"])
                l_wall.append(block["L"])
    except KeyError:
        pass

    wall_dict = {"r_wall": r_wall, "l_wall": l_wall}
    return wall_dict


def make_bound_from_topo(topo_dict):
    BoundaryNames = []
    BoundaryType = []
    BoundaryRmin = []
    BoundaryRmax = []
    BoundaryLmin = []
    BoundaryLmax = []

    for boundary in topo_dict["Boundary"]:
        BoundaryNames.append(boundary)
        tmp_bound_type = []
        tmp_rmin = []
        tmp_rmax = []
        tmp_lmin = []
        tmp_lmax = []
        for bound_element in topo_dict["Boundary"][boundary]:
            tmp_bound_type.append(bound_element["type"])
            tmp_rmin.append(bound_element["Rmin"])
            tmp_rmax.append(bound_element["Rmax"])
            tmp_lmin.append(bound_element["Lmin"])
            tmp_lmax.append(bound_element["Lmax"])
        BoundaryType.append(tmp_bound_type)
        BoundaryRmin.append(tmp_rmin)
        BoundaryRmax.append(tmp_rmax)
        BoundaryLmin.append(tmp_lmin)
        BoundaryLmax.append(tmp_lmax)

    return {
        "names": BoundaryNames,
        "types": BoundaryType,
        "rmin": BoundaryRmin,
        "rmax": BoundaryRmax,
        "lmin": BoundaryLmin,
        "lmax": BoundaryLmax,
    }


def stretch_fun(G, N):
    result = (1.0 - G) / (G * (1 - np.power(G, 1.0 / N)))
    return result


# def stretch_fun(G,N1):
#    result = (1.0-G**(N1/(N1-1)))/((G**(1.0+1.0/(N1-1)))*(1-np.power(G,1.0/(N1-1))))
#    return result


def is_wall(l_wall: list[int], r_wall: list[int], ir: int, il: int) -> int:
    """
    Is the present block a wall (not meshed)

    Parameters
    ----------
    l_wall: list[int]
        List of vertical coordinate index of wall blocks
    r_wall: list[int]
        List of radial coordinate index of wall blocks
    ir: int
        Radial coordinate index of the present block
    il: int
        Vertical coordinate index of the present block

    Returns
    ----------
    is_wall: int
        1 if block (ir,il) is a wall
        0 if block (ir, il) is fluid (and should be meshed)
    """
    is_wall = 0
    for iw in range(len(l_wall)):
        if l_wall[iw] == (il + 1) and r_wall[iw] == ir + 1:
            is_wall = 1
    return is_wall


def merge_and_sort(
    coord_list: list[float], reverse_coord: bool
) -> list[float]:
    """
    Preprocess coordinates to obtain an ordered set of value ameable to block cylindrical meshing

    Parameters
    ----------
    coord_list: list[float]
        List of coordinates
    reverse_coord: bool
        Whether or not to reverse the coordinates

    Returns
    ----------
    clean_list: list[float]
        preprocessed list

    """
    clean_list = list(set(coord_list))
    clean_list.sort(reverse=reverse_coord)
    return clean_list


def verticalCoarsening(
    ratio_properties, ref_block, NVert, L=None, gradVert=None, smooth=False
):
    if gradVert is None:
        gradVert = [1.0 for _ in range(len(NVert))]

    ratio_list = [entry["ratio"] for entry in ratio_properties]
    ratio_dir = []
    ratio_dir_ref = []
    for entry in ratio_properties:
        try:
            ratio_dir.append(entry["direction"])
        except KeyError:
            ratio_dir.append("+")
        try:
            ratio_dir_ref.append(entry["directionRef"])
        except KeyError:
            ratio_dir_ref.append("+")

    for iratio, ratio in enumerate(ratio_list):
        if abs(ratio - 1) < 1e-12:
            pass
        else:
            NVert[iratio] = max(int(round(NVert[iratio] * ratio)), 1)

    block_length = [abs(L[i] - L[i + 1]) for i in range(len(NVert))]
    block_cell_plus_length = [
        block_length[i] / NVert[i] for i in range(len(NVert))
    ]
    block_cell_minus_length = [
        block_length[i] / NVert[i] for i in range(len(NVert))
    ]

    if smooth:
        # Find which block to leave untouched
        indRef = int(ref_block)
        # Decide the order to follow
        indList = []
        for i in range(1, indRef + 1):
            indList.append(indRef - i)
        for i in range(indRef + 1, len(NVert)):
            indList.append(i)

        historyRatio = []
        for ind in indList:
            historyRatio.append(ratio_list[ind])
            if (
                abs(max(historyRatio) - 1) < 1e-12
                and abs(min(historyRatio) - 1) < 1e-12
            ):
                continue

            length = block_length[ind]

            if ratio_dir_ref[ind] == "-":
                deltaE = block_cell_plus_length[ind + 1]
            elif ratio_dir_ref[ind] == "+":
                deltaE = block_cell_minus_length[ind - 1]

            if ratio_dir[ind] == ratio_dir_ref[ind]:
                error_msg = (
                    f"Invalid coarsening ratio for vertical block {ind}"
                )
                error_msg += "\nratio dir and ratio dir ref must be opposite"
                logger.error(error_msg)
                raise ValueError(error_msg)

            if ratio_dir[ind] == "+":
                nl_func = lambda x: stretch_fun(x, NVert[ind])
                gradVert[ind] = 1.0 / bissection(length / deltaE, nl_func)
                iterate = False
                origNVert = NVert[ind]
                while gradVert[ind] < 1 and NVert[ind] > 1:
                    iterate = True
                    NVert[ind] = max(
                        int(round(min(0.99 * NVert[ind], NVert[ind] - 1))), 1
                    )
                    nl_func = lambda x: stretch_fun(x, NVert[ind])
                    gradVert[ind] = 1.0 / bissection(length / deltaE, nl_func)
                if iterate:
                    logger.warning(
                        f"reduced NVert[{ind}] from {origNVert} to {NVert[ind]}"
                    )
                block_cell_minus_length[ind] = deltaE
                block_cell_plus_length[ind] = deltaE * gradVert[ind]

            elif ratio_dir[ind] == "-":
                deltaE = block_cell_minus_length[ind - 1]
                nl_func = lambda x: stretch_fun(x, NVert[ind])
                gradVert[ind] = bissection(length / deltaE, nl_func)

                iterate = False
                origNVert = NVert[ind]
                while gradVert[ind] > 1 and NVert[ind] > 1:
                    iterate = True
                    NVert[ind] = max(
                        int(round(min(0.99 * NVert[ind], NVert[ind] - 1))), 1
                    )
                    nl_func = lambda x: stretch_fun(x, NVert[ind])
                    gradVert[ind] = bissection(length / deltaE, nl_func)
                if iterate:
                    logger.warning(
                        f"reduced NVert[{ind}] from {origNVert} to {NVert[ind]}"
                    )
                block_cell_minus_length[ind] = deltaE / gradVert[ind]
                block_cell_plus_length[ind] = deltaE

    minCell = np.amin(block_cell_minus_length)
    minCell = min(minCell, np.amin(block_cell_plus_length))
    maxCell = np.amax(block_cell_minus_length)
    maxCell = max(maxCell, np.amax(block_cell_plus_length))

    return NVert, gradVert, minCell, maxCell


def radialCoarsening(
    ratio_properties, ref_block, NR, R=None, gradR=None, smooth=False
):
    if gradR is None:
        gradR = [1.0 for _ in range(len(NR))]

    ratio_list = [entry["ratio"] for entry in ratio_properties]
    ratio_dir = []
    ratio_dir_ref = []
    for entry in ratio_properties:
        try:
            ratio_dir.append(entry["direction"])
        except KeyError:
            ratio_dir.append("+")
        try:
            ratio_dir_ref.append(entry["directionRef"])
        except KeyError:
            ratio_dir_ref.append("+")

    for iratio, ratio in enumerate(ratio_list):
        if abs(ratio - 1) < 1e-12:
            pass
        else:
            NR[iratio] = max(int(round(NR[iratio] * ratio)), 1)

    block_length = [R[0] / 2] + [
        abs(R[i] - R[i + 1]) for i in range(len(R) - 1)
    ]
    block_cell_plus_length = [block_length[i] / NR[i] for i in range(len(NR))]
    block_cell_minus_length = [block_length[i] / NR[i] for i in range(len(NR))]

    if smooth:
        # Find which block to leave untouched
        indRef = int(ref_block)
        # Decide the order to follow
        indList = []
        for i in range(1, indRef + 1):
            indList.append(indRef - i)
        for i in range(indRef + 1, len(NR)):
            indList.append(i)

        historyRatio = []
        for ind in indList:
            historyRatio.append(ratio_list[ind])
            if (
                abs(max(historyRatio) - 1) < 1e-12
                and abs(min(historyRatio) - 1) < 1e-12
            ):
                continue
            length = block_length[ind]

            if ratio_dir_ref[ind] == "-":
                deltaE = block_cell_plus_length[ind - 1]
            elif ratio_dir_ref[ind] == "+":
                deltaE = block_cell_minus_length[ind + 1]

            if ratio_dir[ind] == ratio_dir_ref[ind]:
                error_msg = f"Invalid coarsening ratio for radial block {ind}"
                error_msg += "\nratio dir and ratio dir ref must be opposite"
                logger.error(error_msg)
                raise ValueError(error_msg)

            if ratio_dir[ind] == "+":
                nl_func = lambda x: stretch_fun(x, NR[ind])
                gradR[ind] = 1.0 / bissection(length / deltaE, nl_func)
                iterate = False
                origNR = NR[ind]
                while gradR[ind] < 1 and NR[ind] > 1:
                    iterate = True
                    NR[ind] = max(
                        int(round(min(0.99 * NR[ind], NR[ind] - 1))), 1
                    )
                    nl_func = lambda x: stretch_fun(x, NR[ind])
                    gradR[ind] = 1.0 / bissection(length / deltaE, nl_func)
                if iterate:
                    logger.warning(
                        f"reduced NR[{ind}] from {origNR} to {NR[ind]}"
                    )
                block_cell_minus_length[ind] = deltaE
                block_cell_plus_length[ind] = deltaE * gradR[ind]

            elif ratio_dir[ind] == "-":
                deltaE = block_cell_minus_length[ind - 1]
                nl_func = lambda x: stretch_fun(x, NR[ind])
                gradR[ind] = bissection(length / deltaE, nl_func)
                iterate = False
                origNR = NR[ind]
                while gradR[ind] > 1 and NR[ind] > 1:
                    iterate = True
                    NR[ind] = max(int(round(min(0.99 * NR[ind], -1))), 1)
                    nl_func = lambda x: stretch_fun(x, NR[ind])
                    gradR[ind] = bissection(length / deltaE, nl_func)
                if iterate:
                    logger.warning(
                        f"reduced NR[{ind}] from {origNR} to {NR[ind]}"
                    )
                block_cell_minus_length[ind] = deltaE / gradR[ind]
                block_cell_plus_length[ind] = deltaE

    minCell = np.amin(block_cell_minus_length)
    minCell = min(minCell, np.amin(block_cell_plus_length))
    maxCell = np.amax(block_cell_minus_length)
    maxCell = min(maxCell, np.amax(block_cell_plus_length))

    # if smooth:
    #    if gradR is None or R is None:
    #        sys.exit(
    #            "ERROR: cannot smooth radial transition without grading list"
    #        )

    #    Length = R[last_R] - R[last_R - 1]
    #    deltaE = ((R[last_R - 1] - R[last_R - 2])) / NR[last_R - 1]
    #    gradR[last_R] = 1 / (
    #        bissection(Length / deltaE, stretch_fun, NR[last_R])
    #    )
    #    if (gradR[last_R] > 2 or gradR[last_R] < 0.5) and abs(
    #        ratio - 1
    #    ) <= 1e-12:
    #        logger.warning(
    #            "radial smoothing had to be used because your mesh is very coarse"
    #        )
    #        logger.warning("\tIncrease NS in input file to avoid this warning")

    return NR, gradR, minCell, maxCell



import os
import pickle
import shutil
from pathlib import Path

import numpy as np

from bird import BIRD_CASE_DIR, logger
from bird.preprocess.json_gen.design_io import *


def id2simfolder(sim_id: int) -> str:
    """
    Generates simulation folder name from simulation index

    Parameters
    ----------
    sim_id: int
        Simulation index

    Returns
    ----------
    sim_folder : str
        Simulation folder name
    """
    sim_folder = f"Sim_{sim_id:04}"
    return sim_folder


def compare_config(config1, config2):
    same = True
    for key in config1:
        if np.linalg.norm(config1[key] - config2[key]) > 1e-6:
            same = False
            return same
    return same


def check_config(config):
    success = False
    inlet_exist = False
    for key in config:
        if len(np.argwhere(config[key] == 1)) > 0:
            inlet_exist = True
            break
    if inlet_exist:
        success = True
    else:
        success = False
    return success


def save_config_dict(filename, config_dict):
    with open(filename, "wb") as f:
        pickle.dump(config_dict, f)


def load_config_dict(filename):
    with open(filename, "rb") as f:
        config_dict = pickle.load(f)
    return config_dict


def write_script_start(filename, n):
    with open(filename, "w+") as f:
        for i in range(n):
            sim_folder = id2simfolder(i)
            f.write(f"cd {sim_folder}\n")
            f.write(f"sbatch script\n")
            f.write(f"cd ..\n")


def write_script_post(filename, n):
    with open(filename, "w+") as f:
        for i in range(n):
            sim_folder = id2simfolder(i)
            f.write(f"cd {sim_folder}\n")
            f.write(f"sbatch script_post\n")
            f.write(f"cd ..\n")


def write_prep(filename, n):
    with open(filename, "w+") as f:
        f.write("prep () {\n")
        f.write(f"\tcd $1\n")
        f.write(f"\treconstructPar -newTimes\n")
        f.write(f"\tcd ..\n")
        f.write("}\n")
        f.write(f"\n")
        f.write(
            f"source /projects/gas2fuels/ofoam_cray_mpich/OpenFOAM-dev/etc/bashrc\n"
        )
        for i in range(n):
            sim_folder = id2simfolder(i)
            f.write(f"prep {sim_folder}\n")


def overwrite_vvm(case_folder, vvm):
    list_dir = os.listdir(case_folder)
    if not "constant" in list_dir:
        error_msg = f"{case_folder} is likely not a case folder, could not find constant/"
        logger.error(error_msg)
        raise FileNotFoundError(error_msg)
    else:
        filename = os.path.join(case_folder, "constant", "globalVars_temp")
        filename_write = os.path.join(
            case_folder, "constant", "globalVars_temp2"
        )
        with open(filename, "r+") as f:
            lines = f.readlines()
        with open(filename_write, "w+") as f:
            for line in lines:
                if line.startswith("VVM"):
                    f.write(f"VVM\t{vvm};\n")
                else:
                    f.write(line)
        shutil.copy(
            os.path.join(case_folder, "constant", "globalVars_temp2"),
            os.path.join(case_folder, "constant", "globalVars_temp"),
        )
        os.remove(os.path.join(case_folder, "constant", "globalVars_temp2"))


def overwrite_bubble_size_model(case_folder, constantD=False):
    list_dir = os.listdir(case_folder)
    if not "constant" in list_dir:
        error_msg = f"{case_folder} is likely not a case folder, could not find constant/"
        logger.error(error_msg)
        raise FileNotFoundError(error_msg)
    else:
        filename = os.path.join(case_folder, "presteps.sh")
        filename_write = os.path.join(case_folder, "presteps2.sh")
        with open(filename, "r+") as f:
            lines = f.readlines()
        with open(filename_write, "w+") as f:
            for line in lines:
                if line.startswith("cp constant/phaseProperties"):
                    if constantD:
                        f.write(
                            "cp constant/phaseProperties_constantd constant/phaseProperties\n"
                        )
                    else:
                        f.write(
                            "cp constant/phaseProperties_pbe constant/phaseProperties\n"
                        )
                else:
                    f.write(line)
        shutil.copy(
            os.path.join(case_folder, "presteps2.sh"),
            os.path.join(case_folder, "presteps.sh"),
        )
        os.remove(os.path.join(case_folder, "presteps2.sh"))


def generate_small_reactor_cases(
    config_dict,
    branchcom_spots,
    vvm,
    power,
    constantD,
    study_folder,
    template_folder="loop_reactor_pbe_dynmix_nonstat_headbranch",
):
    if not os.path.isabs(template_folder):

        template_folder = os.path.join(
            f"{BIRD_CASE_DIR}", f"{template_folder}"
        )

    geom_dict = make_default_geom_dict_from_file(
        os.path.join(f"{template_folder}", "system", "mesh.json"),
        rescale=0.05,
    )
    try:
        shutil.rmtree(study_folder)
    except:
        pass
    Path(study_folder).mkdir(parents=True, exist_ok=True)
    for sim_id in config_dict:
        sim_folder = id2simfolder(sim_id)
        shutil.copytree(
            f"{template_folder}",
            os.path.join(f"{study_folder}", sim_folder),
        )
        bc_dict = {}
        bc_dict["inlets"] = []
        bc_dict["outlets"] = []
        bc_dict["outlets"].append(
            {
                "branch_id": 6,
                "type": "circle",
                "frac_space": 1,
                "normal_dir": 1,
                "radius": 0.4,
                "nelements": 50,
                "block_pos": "top",
            }
        )
        bc_dict["outlets"].append(
            {
                "branch_id": 4,
                "type": "circle",
                "frac_space": 1,
                "normal_dir": 1,
                "radius": 0.4,
                "nelements": 50,
                "block_pos": "top",
            }
        )
        for branch in config_dict:
            if branch in [0, 1, 2]:
                ind = np.argwhere(config_dict[sim_id][branch] == 1)
                if len(ind) > 0:
                    ind = list(ind[:, 0])
                    for iind in ind:
                        bc_dict["inlets"].append(
                            {
                                "branch_id": branch,
                                "type": "circle",
                                "frac_space": branchcom_spots[branch][iind],
                                "normal_dir": 1,
                                "radius": 0.4,
                                "nelements": 50,
                                "block_pos": "bottom",
                            }
                        )
        generate_stl_patch(
            os.path.join(
                study_folder, sim_folder, "system", "inlets_outlets.json"
            ),
            bc_dict,
            geom_dict,
        )

        mix_list = []
        for branch in config_dict:
            if branch in [0, 1, 2]:
                ind = np.argwhere(config_dict[sim_id][branch] == 0)
                if len(ind) > 0:
                    ind = list(ind[:, 0])
                    for iind in ind:
                        if branch == 0:
                            sign = "+"
                        else:
                            sign = "-"
                        mix_list.append(
                            {
                                "branch_id": branch,
                                "frac_space": branchcom_spots[branch][iind],
                                "start_time": 1,
                                "power": power,
                                "sign": sign,
                            }
                        )
        generate_dynamic_mixer(
            os.path.join(study_folder, sim_folder, "system", "mixers.json"),
            mix_list,
            geom_dict,
        )
        overwrite_vvm(
            case_folder=os.path.join(study_folder, sim_folder), vvm=vvm
        )
        overwrite_bubble_size_model(
            case_folder=os.path.join(study_folder, sim_folder),
            constantD=constantD,
        )

    geom_dict = make_default_geom_dict_from_file(
        os.path.join(f"{template_folder}", "system", "mesh.json"),
        rescale=0.05,
    )


def generate_scaledup_reactor_cases(
    config_dict,
    branchcom_spots,
    vvm,
    power,
    constantD,
    study_folder,
    template_folder="loop_reactor_pbe_dynmix_nonstat_headbranch_scaleup",
):

    if not os.path.isabs(template_folder):
        template_folder = os.path.join(
            f"{BIRD_CASE_DIR}", f"{template_folder}"
        )

    geom_dict = make_default_geom_dict_from_file(
        os.path.join(f"{template_folder}", "system", "mesh.json")
    )
    try:
        shutil.rmtree(study_folder)
    except:
        pass
    Path(study_folder).mkdir(parents=True, exist_ok=True)
    for sim_id in config_dict:
        sim_folder = id2simfolder(sim_id)
        shutil.copytree(
            f"{template_folder}",
            os.path.join(f"{study_folder}", sim_folder),
        )
        bc_dict = {}
        bc_dict["inlets"] = []
        bc_dict["outlets"] = []
        bc_dict["outlets"].append(
            {
                "branch_id": 6,
                "type": "circle",
                "frac_space": 1,
                "normal_dir": 1,
                "radius": 0.4,
                "nelements": 50,
                "block_pos": "top",
            }
        )
        bc_dict["outlets"].append(
            {
                "branch_id": 4,
                "type": "circle",
                "frac_space": 1,
                "normal_dir": 1,
                "radius": 0.4,
                "nelements": 50,
                "block_pos": "top",
            }
        )
        for branch in config_dict:
            if branch in [0, 1, 2]:
                ind = np.argwhere(config_dict[sim_id][branch] == 1)
                if len(ind) > 0:
                    ind = list(ind[:, 0])
                    for iind in ind:
                        bc_dict["inlets"].append(
                            {
                                "branch_id": branch,
                                "type": "circle",
                                "frac_space": branchcom_spots[branch][iind],
                                "normal_dir": 1,
                                "radius": 0.4,
                                "nelements": 50,
                                "block_pos": "bottom",
                            }
                        )
        generate_stl_patch(
            os.path.join(
                study_folder, sim_folder, "system", "inlets_outlets.json"
            ),
            bc_dict,
            geom_dict,
        )

        mix_list = []
        for branch in config_dict:
            if branch in [0, 1, 2]:
                ind = np.argwhere(config_dict[sim_id][branch] == 0)
                if len(ind) > 0:
                    ind = list(ind[:, 0])
                    for iind in ind:
                        if branch == 0:
                            sign = "+"
                        else:
                            sign = "-"
                        mix_list.append(
                            {
                                "branch_id": branch,
                                "frac_space": branchcom_spots[branch][iind],
                                "start_time": 3,
                                "power": power,
                                "sign": sign,
                            }
                        )
        generate_dynamic_mixer(
            os.path.join(study_folder, sim_folder, "system", "mixers.json"),
            mix_list,
            geom_dict,
        )
        overwrite_vvm(
            case_folder=os.path.join(study_folder, sim_folder), vvm=vvm
        )
        overwrite_bubble_size_model(
            case_folder=os.path.join(study_folder, sim_folder),
            constantD=constantD,
        )


def check_sparger_config(
    sparger_locs: list[float],
    n_spargers: int | None,
    sparger_spacing: float,
    edge_spacing: float,
    n_branches: int,
    bypass_sparger_spacing: bool,
) -> None:
    """
    Check realizability of the sparger placement configuration

    Parameters
    ----------
    sparger_locs : list[float]
        Location of every sparger along the loop reactor coordinate [-]
        There are 3 branches. Spargers can be placed anywhere
        between edge_spacing and (1-edge_spacing) fractions of the branch
        Each sparger locations must be between 0 and 3*1=3
    n_spargers : int|None
        Number of spargers
    sparger_spacing : float
        Spacing between two spargers [-]
    edge_spacing : float
        Spacing required between any sparger and the edges of the branches [-]
    n_branches : int
        Number of loop reactor branches
    bypass_sparger_spacing: bool
        If true, allow an overlap of spargers
    """

    # Check that number of spargers is consistent
    if n_spargers is None:
        n_spargers = len(sparger_locs)
    else:
        assert n_spargers == len(sparger_locs)
    assert n_spargers >= 1

    # Basis check on the number of branches
    assert n_branches > 0

    # Check that locations of spargers is consistent
    # There are n_branches branches. Spargers can be placed anywhere
    # between edge_spacing and (1-edge_spacing) fractions of the branch
    # Each sparger locations must be between 0 and n_branches*1=n_branches
    assert edge_spacing > 0
    assert edge_spacing < 1
    assert all(np.array(sparger_locs) >= 0)
    assert all(np.array(sparger_locs) <= float(n_branches))
    for ibranch in range(n_branches):
        if ibranch == 0:
            assert not np.any(np.array(sparger_locs) < edge_spacing)
        if ibranch == n_branches - 1:
            assert not np.any(
                np.array(sparger_locs) > float(n_branches) - edge_spacing
            )
        assert not np.any(
            (np.array(sparger_locs) > float(ibranch) + 1.0 - edge_spacing)
            & (np.array(sparger_locs) < float(ibranch) + 1.0 + edge_spacing)
        )

    # Check that spargers are sufficiently spaced out
    assert sparger_spacing >= 0
    if not bypass_sparger_spacing:
        assert all(np.diff(np.sort(np.array(sparger_locs))) >= sparger_spacing)


def generate_single_scaledup_reactor_sparger_cases(
    sparger_locs: list[float],
    n_spargers: int | None = None,
    sparger_spacing: float = 0.15,
    edge_spacing: float = 0.2,
    n_branches: int = 3,
    sim_id: int = 0,
    constantD: bool = True,
    vvm: float = 0.4,
    study_folder: str = ".",
    template_folder: str = "loop_reactor_pbe_dynmix_nonstat_headbranch_scaleup",
    bypass_sparger_spacing: bool = False,
):
    """
    Generates loop reactor case with desired sparger placement configuration

    Parameters
    ----------
    sparger_locs : list[float]
        Location of every sparger along the loop reactor coordinate [-]
    n_spargers : int|None
        Number of spargers
    sparger_spacing : float
        Spacing between two spargers [-]
    edge_spacing : float
        Spacing required between any sparger and the edges of the branches [-]
    n_branches : int
        Number of loop reactor branches
    sim_id : int
        Index identifier of the simulation
    constantD : bool
        If true, use constant bubble diameter
        If false, use population balance
    vvm : float
        VVM value [-]
    study_folder : str
        Where to generate the case
    template_folder: str
        The case template to start from
    bypass_sparger_spacing: bool
        If true, allow an overlap of spargers
    """

    # Sanity checks
    check_sparger_config(
        sparger_locs=sparger_locs,
        n_spargers=n_spargers,
        sparger_spacing=sparger_spacing,
        edge_spacing=edge_spacing,
        n_branches=n_branches,
        bypass_sparger_spacing=bypass_sparger_spacing,
    )

    # Find on which branch is each sparger
    branch_id = [int(entry) for entry in sparger_locs]

    # Case generation
    if not os.path.isabs(template_folder):

        template_folder = os.path.join(
            f"{BIRD_CASE_DIR}", f"{template_folder}"
        )
    geom_dict = make_default_geom_dict_from_file(
        os.path.join(f"{template_folder}", "system", "mesh.json")
    )

    # Start from template
    sim_folder = id2simfolder(sim_id)
    shutil.copytree(
        f"{template_folder}",
        os.path.join(f"{study_folder}", sim_folder),
    )

    bc_dict = {}
    bc_dict["inlets"] = []
    bc_dict["outlets"] = []
    bc_dict["outlets"].append(
        {
            "branch_id": 6,
            "type": "circle",
            "frac_space": 1,
            "normal_dir": 1,
            "radius": 0.4,
            "nelements": 50,
            "block_pos": "top",
        }
    )
    bc_dict["outlets"].append(
        {
            "branch_id": 4,
            "type": "circle",
            "frac_space": 1,
            "normal_dir": 1,
            "radius": 0.4,
            "nelements": 50,
            "block_pos": "top",
        }
    )

    for branch, loc in zip(branch_id, sparger_locs):
        bc_dict["inlets"].append(
            {
                "branch_id": branch,
                "type": "circle",
                "frac_space": loc - branch,
                "normal_dir": 1,
                "radius": 0.4,
                "nelements": 50,
                "block_pos": "bottom",
            }
        )

    generate_stl_patch(
        os.path.join(
            study_folder, sim_folder, "system", "inlets_outlets.json"
        ),
        bc_dict,
        geom_dict,
    )

    mix_list = []
    generate_dynamic_mixer(
        os.path.join(study_folder, sim_folder, "system", "mixers.json"),
        mix_list,
        geom_dict,
    )
    overwrite_vvm(case_folder=os.path.join(study_folder, sim_folder), vvm=vvm)
    overwrite_bubble_size_model(
        case_folder=os.path.join(study_folder, sim_folder),
        constantD=constantD,
    )



# from reactor_geom_data import *
import os

import numpy as np

from bird.meshing._stirred_tank_reactor import StirredTankReactor


def get_reactor_geom(yamlfile):
    return StirredTankReactor.from_file(yamlfile)


def write_ofoam_preamble(outfile, react):
    outfile.write(
        "/*--------------------------------*- C++ -*----------------------------------*\\\n"
    )
    outfile.write(
        "| =========                 |                                                 |\n"
    )
    outfile.write(
        "| \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox           |\n"
    )
    outfile.write(
        "|  \\    /   O peration     | Version:  5                                     |\n"
    )
    outfile.write(
        "|   \\  /    A nd           | Web:      www.OpenFOAM.org                      |\n"
    )
    outfile.write(
        "|    \\/     M anipulation  |                                                 |\n"
    )
    outfile.write(
        "\*---------------------------------------------------------------------------*/\n"
    )
    outfile.write("FoamFile\n")
    outfile.write("{\n")
    outfile.write("\tversion     2.0;\n")
    outfile.write("\tformat      ascii;\n")
    outfile.write("\tclass       dictionary;\n")
    outfile.write("\tobject      blockMeshDict;\n")
    outfile.write("}\n\n")
    outfile.write(
        "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n\n"
    )
    outfile.write("convertToMeters 1.0;\n\n")
    outfile.write(
        "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n\n"
    )


def write_vertices(outfile, react):
    outfile.write(
        "\n// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
    )
    outfile.write("vertices\n(\n")

    nsections = react.nsections
    reacthts = react.reacthts
    dangle = react.dangle
    nsplits = react.nsplits
    ncirc = react.ncirc
    circradii = react.circradii
    polyrad = react.polyrad

    counter = 0
    for repeat in range(2):
        for zi in range(nsections):
            outfile.write("\n//section " + str(zi) + "\n")
            outfile.write("\n//center\n")
            outfile.write(
                "(0.0 0.0 " + str(reacthts[zi]) + ") // " + str(counter) + "\n"
            )
            counter = counter + 1

            # polygon section
            outfile.write("\n//polygon\n")
            for i in range(nsplits):
                ang = i * dangle
                x = polyrad * np.cos(ang)
                y = polyrad * np.sin(ang)
                outfile.write(
                    "( "
                    + str(x)
                    + " "
                    + str(y)
                    + " "
                    + str(reacthts[zi])
                    + " ) // "
                    + str(counter)
                    + "\n"
                )
                counter = counter + 1

            outfile.write("\n//circles\n")

            for ci in range(ncirc):
                outfile.write("\n//circle " + str(ci) + "\n")

                for i in range(nsplits):
                    ang = i * dangle
                    x = circradii[ci] * np.cos(ang)
                    y = circradii[ci] * np.sin(ang)
                    outfile.write(
                        "( "
                        + str(x)
                        + " "
                        + str(y)
                        + " "
                        + str(reacthts[zi])
                        + " ) //"
                        + str(counter)
                        + "\n"
                    )
                    counter = counter + 1

    outfile.write(");\n")


def get_globalindex_of(splti, ci, zi, react):
    npts_per_section = react.npts_per_section
    centeroffset = react.centeroffset
    polyoffset = react.polyoffset
    nsplits = react.nsplits
    hub_circ = react.hub_circ
    tank_circ = react.tank_circ

    # also works for ci=-1
    global_id = (
        zi * npts_per_section
        + centeroffset
        + polyoffset
        + ci * nsplits
        + splti % nsplits
    )
    return global_id


def get_baffle_point_of(splti, ci, zi, react):
    baff_sections = react.baff_sections
    nsections = react.nsections
    npts_per_section = react.npts_per_section
    hub_circ = react.hub_circ
    tank_circ = react.tank_circ

    baffle_id = get_globalindex_of(splti, ci, zi, react)

    if zi in baff_sections:
        if ci == hub_circ:
            if splti % 2 == 0:  # even number for impeller
                baffle_id += nsections * npts_per_section

        if ci == tank_circ:
            if splti % 2 == 1:  # odd number for baffle
                baffle_id += nsections * npts_per_section

    return baffle_id


def write_edges(outfile, react):
    nsections = react.nsections
    nsplits = react.nsplits
    dangle = react.dangle
    circradii = react.circradii
    ncirc = react.ncirc
    reacthts = react.reacthts

    outfile.write(
        "\n// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
    )
    outfile.write("edges\n(\n")

    for zi in range(nsections):
        outfile.write("\n//section " + str(zi) + "\n")

        offset = 1 + nsplits  # one for center and nsplits for polygon

        outfile.write("\n//circles\n")
        for ci in range(ncirc):
            outfile.write("\n//circle " + str(ci) + "\n")
            for i in range(nsplits):
                ang = i * dangle
                midx = circradii[ci] * np.cos(ang + dangle / 2)
                midy = circradii[ci] * np.sin(ang + dangle / 2)

                globalind1 = get_baffle_point_of(i, ci, zi, react)
                globalind2 = get_globalindex_of(i + 1, ci, zi, react)

                outfile.write(
                    "arc " + str(globalind1) + " " + str(globalind2) + " "
                )
                outfile.write(
                    "( "
                    + str(midx)
                    + " "
                    + str(midy)
                    + " "
                    + str(reacthts[zi])
                    + " )\n"
                )

    outfile.write(");\n")


def write_this_block(outfile, comment, ids, mesh, zonename="none"):
    outfile.write("\n //" + comment + "\n")
    outfile.write("hex (")
    for i in range(len(ids)):
        outfile.write(str(ids[i]) + " ")
    outfile.write(")\n")

    if zonename != "none":
        outfile.write(zonename + "\n")

    outfile.write("( %d %d %d )\n" % (mesh[0], mesh[1], mesh[2]))
    outfile.write("SimpleGrading (1 1 1)\n")


def write_blocks(outfile, react):
    outfile.write(
        "\n// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
    )
    outfile.write("blocks\n(\n")

    idarray = np.zeros(8, dtype=int)
    mesharray = np.zeros(3, dtype=int)

    npts_per_section = react.npts_per_section
    nonstem_volumes = react.nonstem_volumes
    nsplits = react.nsplits
    centeroffset = react.centeroffset
    Na = react.Na
    Npoly = react.Npoly
    meshz = react.meshz
    meshr = react.meshr
    ncirc = react.ncirc
    inhub_circ = react.inhub_circ
    hub_circ = react.hub_circ
    mrf_circ = react.mrf_circ
    nvolumes = react.nvolumes
    mrf_volumes = react.mrf_volumes
    hub_volumes = react.hub_volumes

    for zi in range(nvolumes):
        outfile.write("\n//section " + str(zi) + "-" + str(zi + 1) + "\n")

        offset0 = zi * npts_per_section
        offset1 = (zi + 1) * npts_per_section

        # polygon section
        center_id0 = offset0
        center_id1 = offset1

        # skip polygon blocks in stem sections
        if zi in nonstem_volumes:
            for i in range(nsplits):
                localind1 = centeroffset + i % nsplits
                localind2 = centeroffset + (i + 1) % nsplits

                idarray[0] = offset0 + localind1
                idarray[1] = offset1 + localind1
                idarray[2] = offset1 + localind2
                idarray[3] = offset0 + localind2
                idarray[4] = center_id0
                idarray[5] = center_id1
                idarray[6] = center_id1
                idarray[7] = center_id0

                mesharray[0] = meshz[zi]
                mesharray[1] = Na
                mesharray[2] = Npoly

                zonename = "none"
                if zi in mrf_volumes:
                    zonename = "rotor"

                write_this_block(
                    outfile, "block %d" % (i), idarray, mesharray, zonename
                )

        idarray[:] = 0
        mesharray[:] = 0
        outfile.write("\n//circles\n")

        for ci in range(ncirc):
            zonename = "none"

            # skip blocks inside hub
            if ((ci == inhub_circ) or (ci == hub_circ)) and (
                zi in hub_volumes
            ):
                continue

            outfile.write("\n//circle " + str(ci) + "\n")

            if (zi in mrf_volumes) and (ci <= mrf_circ):
                zonename = "rotor"

            for i in range(nsplits):
                idarray[0] = get_baffle_point_of(i, ci, zi, react)
                idarray[1] = get_baffle_point_of(i, ci, zi + 1, react)
                idarray[2] = get_globalindex_of(i + 1, ci, zi + 1, react)
                idarray[3] = get_globalindex_of(i + 1, ci, zi, react)
                idarray[4] = get_baffle_point_of(i, ci - 1, zi, react)
                idarray[5] = get_baffle_point_of(i, ci - 1, zi + 1, react)
                idarray[6] = get_globalindex_of(i + 1, ci - 1, zi + 1, react)
                idarray[7] = get_globalindex_of(i + 1, ci - 1, zi, react)

                mesharray[0] = meshz[zi]
                mesharray[1] = Na
                mesharray[2] = meshr[ci]
                write_this_block(
                    outfile, "block %d" % (i), idarray, mesharray, zonename
                )

    outfile.write(");\n")


def write_patches(outfile, react):
    inhub_ci = react.inhub_circ
    hub_ci = react.hub_circ
    rot_ci = react.rot_circ
    npts_per_section = react.npts_per_section
    nsplits = react.nsplits
    hub_volumes = react.hub_volumes
    nsections = react.nsections
    ncirc = react.ncirc
    nimpellers = react.nimpellers
    baff_volumes = react.baff_volumes
    nonbaff_volumes = react.nonbaff_volumes
    only_stem_volumes = react.only_stem_volumes

    poly_ci = -1

    outfile.write(
        "\n// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
    )
    outfile.write("patches\n(\n")

    # inlet patch
    zi = 0
    centerid = zi * npts_per_section

    outfile.write("\n\tpatch inlet\n\t(\n")

    # polygon
    outfile.write("\n\t\t//polygon\n")
    for i in range(nsplits):
        outfile.write("\t\t( ")
        outfile.write(str(get_globalindex_of(i, poly_ci, zi, react)) + " ")
        outfile.write(str(get_globalindex_of(i + 1, poly_ci, zi, react)) + " ")
        outfile.write(str(centerid) + " ")
        outfile.write(str(centerid) + ")\n")

    outfile.write("\n\t\t//inhub_circ to polygon\n")
    for i in range(nsplits):
        outfile.write("\t\t( ")
        outfile.write(str(get_globalindex_of(i, inhub_ci, zi, react)) + " ")
        outfile.write(
            str(get_globalindex_of(i + 1, inhub_ci, zi, react)) + " "
        )
        outfile.write(str(get_globalindex_of(i + 1, poly_ci, zi, react)) + " ")
        outfile.write(str(get_globalindex_of(i, poly_ci, zi, react)) + ")\n")

    outfile.write("\n\t\t//hub to inhub_circ\n")
    for i in range(nsplits):
        outfile.write("\t\t( ")
        outfile.write(str(get_globalindex_of(i, hub_ci, zi, react)) + " ")
        outfile.write(str(get_globalindex_of(i + 1, hub_ci, zi, react)) + " ")
        outfile.write(
            str(get_globalindex_of(i + 1, inhub_ci, zi, react)) + " "
        )
        outfile.write(str(get_globalindex_of(i, inhub_ci, zi, react)) + ")\n")

    outfile.write("\n\t\t//rotor to hub\n")
    for i in range(nsplits):
        outfile.write("\t\t( ")
        outfile.write(str(get_globalindex_of(i, rot_ci, zi, react)) + " ")
        outfile.write(str(get_globalindex_of(i + 1, rot_ci, zi, react)) + " ")
        outfile.write(str(get_globalindex_of(i + 1, hub_ci, zi, react)) + " ")
        outfile.write(str(get_globalindex_of(i, hub_ci, zi, react)) + ")\n")

    outfile.write("\t)\n")

    # outlet patch
    zi = nsections - 1
    centerid = zi * npts_per_section
    outfile.write("\n\tpatch outlet\n\t(\n")

    # no polygon patch in outlet when we include stem
    # polygon
    # outfile.write("\n\t\t//polygon\n")
    # for i in range(nsplits):
    #    outfile.write("\t\t( ")
    #    outfile.write(str(get_globalindex_of(i,poly_ci,zi))+" ")
    #    outfile.write(str(get_globalindex_of(i+1,poly_ci,zi))+" ")
    #    outfile.write(str(centerid)+" ")
    #    outfile.write(str(centerid)+")\n")

    outfile.write("\n\t\t//circles\n")
    for ci in range(ncirc):
        outfile.write(
            "\n\t\t//circle " + str(ci) + " - " + str(ci - 1) + " \n"
        )
        for i in range(nsplits):
            outfile.write("\t\t( ")
            outfile.write(str(get_globalindex_of(i, ci, zi, react)) + " ")
            outfile.write(str(get_globalindex_of(i + 1, ci, zi, react)) + " ")
            outfile.write(
                str(get_globalindex_of(i + 1, ci - 1, zi, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i, ci - 1, zi, react)) + ")\n"
            )

    outfile.write("\t)\n")

    # propeller patch
    outfile.write("\n\twall propeller\n\t(\n")

    # need polygon patch at the first impeller
    zi = hub_volumes[0]
    outfile.write("\n\t\t//polygon\n")
    centerid = zi * npts_per_section
    # polygon
    for i in range(nsplits):
        outfile.write("\t\t( ")
        outfile.write(str(get_globalindex_of(i, poly_ci, zi, react)) + " ")
        outfile.write(str(get_globalindex_of(i + 1, poly_ci, zi, react)) + " ")
        outfile.write(str(centerid) + " ")
        outfile.write(str(centerid) + ")\n")

    for n_imp in range(nimpellers):
        zi_bottom = hub_volumes[n_imp]  # bottom of impeller section
        zi_top = zi_bottom + 1  # bottom of impeller section

        for zi in [zi_bottom, zi_top]:
            outfile.write("\n\t\t//hub to blade circle\n")
            for i in range(nsplits):
                outfile.write("\t\t( ")
                outfile.write(
                    str(get_baffle_point_of(i, hub_ci, zi, react)) + " "
                )
                outfile.write(
                    str(get_globalindex_of(i + 1, hub_ci, zi, react)) + " "
                )
                outfile.write(
                    str(get_globalindex_of(i + 1, inhub_ci, zi, react)) + " "
                )
                outfile.write(
                    str(get_baffle_point_of(i, inhub_ci, zi, react)) + ")\n"
                )

            outfile.write("\n\t\t//blade to polygon\n")
            for i in range(nsplits):
                outfile.write("\t\t( ")
                outfile.write(
                    str(get_baffle_point_of(i, inhub_ci, zi, react)) + " "
                )
                outfile.write(
                    str(get_globalindex_of(i + 1, inhub_ci, zi, react)) + " "
                )
                outfile.write(
                    str(get_globalindex_of(i + 1, poly_ci, zi, react)) + " "
                )
                outfile.write(
                    str(get_baffle_point_of(i, poly_ci, zi, react)) + ")\n"
                )

        # sides
        outfile.write("\n\t\t//sides\n")
        for i in range(nsplits):
            outfile.write("\t\t( ")
            outfile.write(
                str(get_baffle_point_of(i, hub_ci, zi_bottom, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i + 1, hub_ci, zi_bottom, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i + 1, hub_ci, zi_top, react)) + " "
            )
            outfile.write(
                str(get_baffle_point_of(i, hub_ci, zi_top, react)) + ")\n"
            )

    # blades
    outfile.write("\n\t\t//blades\n")
    for zi in baff_volumes:
        zi_bottom = zi
        zi_top = zi + 1
        for i in range(0, nsplits, 2):  # even numbers
            outfile.write("\t\t( ")
            outfile.write(
                str(get_baffle_point_of(i, hub_ci + 1, zi_bottom, react)) + " "
            )
            outfile.write(
                str(get_baffle_point_of(i, hub_ci + 1, zi_top, react)) + " "
            )
            outfile.write(
                str(get_baffle_point_of(i, hub_ci, zi_top, react)) + " "
            )
            outfile.write(
                str(get_baffle_point_of(i, hub_ci, zi_bottom, react)) + ")\n"
            )

            outfile.write("\t\t( ")
            outfile.write(
                str(get_globalindex_of(i, hub_ci + 1, zi_bottom, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i, hub_ci + 1, zi_top, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i, hub_ci, zi_top, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i, hub_ci, zi_bottom, react)) + ")\n"
            )

    # inside hub blades
    for n_imp in range(nimpellers):
        hub_vol = hub_volumes[n_imp]
        zi_pairs = [[hub_vol - 1, hub_vol], [hub_vol + 1, hub_vol + 2]]

        for zi_pair in zi_pairs:
            zi_below = zi_pair[0]
            zi_above = zi_pair[1]

            for i in range(0, nsplits, 2):  # even numbers
                outfile.write("\t\t( ")
                outfile.write(
                    str(get_baffle_point_of(i, hub_ci, zi_below, react)) + " "
                )
                outfile.write(
                    str(get_baffle_point_of(i, inhub_ci, zi_below, react))
                    + " "
                )
                outfile.write(
                    str(get_baffle_point_of(i, inhub_ci, zi_above, react))
                    + " "
                )
                outfile.write(
                    str(get_baffle_point_of(i, hub_ci, zi_above, react))
                    + ")\n"
                )

                outfile.write("\t\t( ")
                outfile.write(
                    str(get_globalindex_of(i, hub_ci, zi_below, react)) + " "
                )
                outfile.write(
                    str(get_globalindex_of(i, inhub_ci, zi_below, react)) + " "
                )
                outfile.write(
                    str(get_globalindex_of(i, inhub_ci, zi_above, react)) + " "
                )
                outfile.write(
                    str(get_globalindex_of(i, hub_ci, zi_above, react)) + ")\n"
                )

    # stem
    outfile.write("\n\t\t//stem sides\n")
    for zi in only_stem_volumes:
        zi_bottom = zi
        zi_top = zi + 1
        for i in range(nsplits):
            outfile.write("\t\t( ")
            outfile.write(
                str(get_globalindex_of(i, poly_ci, zi_bottom, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i + 1, poly_ci, zi_bottom, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i + 1, poly_ci, zi_top, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i, poly_ci, zi_top, react)) + ")\n"
            )

    outfile.write("\t)\n")

    # stator and walls patch
    tank_ci = ncirc - 1
    outfile.write("\n\twall walls\n\t(\n")

    for zi in range(nsections - 1):
        outfile.write(
            "\n\t\t//tank walls " + str(zi) + " - " + str(zi + 1) + "\n"
        )

        for i in range(nsplits):
            outfile.write("\t\t( ")
            outfile.write(
                str(get_baffle_point_of(i, tank_ci, zi, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i + 1, tank_ci, zi, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i + 1, tank_ci, zi + 1, react)) + " "
            )
            outfile.write(
                str(get_baffle_point_of(i, tank_ci, zi + 1, react)) + ")\n"
            )

        # baffles
        outfile.write("\n\t\t//baffles\n")
        for i in range(1, nsplits, 2):  # all odd numbers
            outfile.write("\t\t( ")
            outfile.write(
                str(get_baffle_point_of(i, tank_ci, zi, react)) + " "
            )
            outfile.write(
                str(get_baffle_point_of(i, tank_ci - 1, zi, react)) + " "
            )
            outfile.write(
                str(get_baffle_point_of(i, tank_ci - 1, zi + 1, react)) + " "
            )
            outfile.write(
                str(get_baffle_point_of(i, tank_ci, zi + 1, react)) + ")\n"
            )

            outfile.write("\t\t( ")
            outfile.write(str(get_globalindex_of(i, tank_ci, zi, react)) + " ")
            outfile.write(
                str(get_globalindex_of(i, tank_ci - 1, zi, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i, tank_ci - 1, zi + 1, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i, tank_ci, zi + 1, react)) + ")\n"
            )

    # inlet wall patch
    # skip rotor to hub to inhub_circ to polygon region
    # which is covered in inflow
    zi = 0
    outfile.write("\n\t\t//circles\n")
    for ci in range(rot_ci + 1, ncirc):  # start from rotor circle
        outfile.write(
            "\n\t\t//circle " + str(ci) + " - " + str(ci - 1) + " \n"
        )
        for i in range(nsplits):
            outfile.write("\t\t( ")
            outfile.write(str(get_globalindex_of(i, ci, zi, react)) + " ")
            outfile.write(str(get_globalindex_of(i + 1, ci, zi, react)) + " ")
            outfile.write(
                str(get_globalindex_of(i + 1, ci - 1, zi, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i, ci - 1, zi, react)) + ")\n"
            )

    outfile.write("\t)\n")

    # stitch faces patches
    outfile.write("\n\tempty inside_to_hub\n\t(\n")

    zi_pairs = []
    for vols in nonbaff_volumes:
        zi_pairs.append([vols, vols + 1])

    for zi_pair in zi_pairs:
        zi_below = zi_pair[0]
        zi_above = zi_pair[1]
        outfile.write(
            "\n\t\t//pair :" + str(zi_below) + "-" + str(zi_above) + "\n"
        )

        for i in range(0, nsplits, 2):  # even numbers
            outfile.write("\t\t( ")
            outfile.write(
                str(get_baffle_point_of(i, hub_ci, zi_below, react)) + " "
            )
            outfile.write(
                str(get_baffle_point_of(i, inhub_ci, zi_below, react)) + " "
            )
            outfile.write(
                str(get_baffle_point_of(i, inhub_ci, zi_above, react)) + " "
            )
            outfile.write(
                str(get_baffle_point_of(i, hub_ci, zi_above, react)) + ")\n"
            )

    outfile.write("\t)\n")

    outfile.write("\n\tempty inside_to_hub_copy\n\t(\n")

    for zi_pair in zi_pairs:
        zi_below = zi_pair[0]
        zi_above = zi_pair[1]
        outfile.write(
            "\n\t\t//pair :" + str(zi_below) + "-" + str(zi_above) + "\n"
        )

        for i in range(0, nsplits, 2):  # even numbers
            outfile.write("\t\t( ")
            outfile.write(
                str(get_globalindex_of(i, hub_ci, zi_below, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i, inhub_ci, zi_below, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i, inhub_ci, zi_above, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i, hub_ci, zi_above, react)) + ")\n"
            )

    outfile.write("\t)\n")

    # stitch faces patches
    outfile.write("\n\tempty hub_to_rotor\n\t(\n")

    for zi_pair in zi_pairs:
        zi_below = zi_pair[0]
        zi_above = zi_pair[1]
        outfile.write(
            "\n\t\t//pair :" + str(zi_below) + "-" + str(zi_above) + "\n"
        )

        for i in range(0, nsplits, 2):  # even numbers
            outfile.write("\t\t( ")
            outfile.write(
                str(get_baffle_point_of(i, rot_ci, zi_below, react)) + " "
            )
            outfile.write(
                str(get_baffle_point_of(i, hub_ci, zi_below, react)) + " "
            )
            outfile.write(
                str(get_baffle_point_of(i, hub_ci, zi_above, react)) + " "
            )
            outfile.write(
                str(get_baffle_point_of(i, rot_ci, zi_above, react)) + ")\n"
            )

    outfile.write("\t)\n")
    outfile.write("\n\tempty hub_to_rotor_copy\n\t(\n")

    for zi_pair in zi_pairs:
        zi_below = zi_pair[0]
        zi_above = zi_pair[1]
        outfile.write(
            "\n\t\t//pair :" + str(zi_below) + "-" + str(zi_above) + "\n"
        )

        for i in range(0, nsplits, 2):  # even numbers
            outfile.write("\t\t( ")
            outfile.write(
                str(get_globalindex_of(i, rot_ci, zi_below, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i, hub_ci, zi_below, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i, hub_ci, zi_above, react)) + " "
            )
            outfile.write(
                str(get_globalindex_of(i, rot_ci, zi_above, react)) + ")\n"
            )

    outfile.write("\t)\n")

    outfile.write(");\n")


if __name__ == "__main__":
    react = get_reactor_geom(
        "stirred_tank_mesh_templates/base_tank/tank_par.yaml"
    )
    with open("tmp", "w") as outfile:
        write_ofoam_preamble(outfile, react)
        write_vertices(outfile, react)
        write_edges(outfile, react)
        write_blocks(outfile, react)
        write_patches(outfile, react)



import os
from pathlib import Path

import numpy as np

from bird.utilities.parser import parse_yaml


class StirredTankReactor:
    def __init__(
        self,
        Dt,
        Da,
        H,
        nimpellers,
        C,
        W,
        L,
        Lin,
        J,
        Wh,
        polyrad,
        Z0,
        nr,
        nz,
        Npoly,
        Na,
        nbaffles,
    ):
        # Loop through params and setattr v to self.k
        for k, v in locals().items():
            if k != "self":
                setattr(self, k, v)
        self.Dh = Da - 2 * L
        self.Dmrf = (Da + Dt - 2 * J) / 2
        self.nsplits = 2 * nbaffles  # we need twice the number of splits
        self.dangle = 2.0 * np.pi / float(self.nsplits)

        self.circradii = np.array(
            [
                self.Dh / 2 - Lin,
                self.Dh / 2,
                Da / 2,
                self.Dmrf / 2,
                Dt / 2 - J,
                Dt / 2,
            ]
        )
        self.ncirc = len(self.circradii)
        self.hub_circ = 1
        self.inhub_circ = self.hub_circ - 1  # circle inside hub
        self.rot_circ = self.hub_circ + 1
        self.mrf_circ = self.rot_circ + 1
        self.tank_circ = self.ncirc - 1

        self.reacthts = [Z0]
        self.baff_sections = []
        self.baff_volumes = []
        self.hub_volumes = []
        count = 1
        for n_imp in range(self.nimpellers):
            self.reacthts.append(Z0 + C[n_imp] - W / 2)

            self.baff_sections.append(count)
            self.baff_volumes.append(count)
            count = count + 1

            self.reacthts.append(Z0 + C[n_imp] - Wh / 2)

            self.baff_sections.append(count)
            self.baff_volumes.append(count)
            self.hub_volumes.append(count)
            count = count + 1

            self.reacthts.append(Z0 + C[n_imp] + Wh / 2)

            self.baff_sections.append(count)
            self.baff_volumes.append(count)
            count = count + 1

            self.reacthts.append(Z0 + C[n_imp] + W / 2)
            self.baff_sections.append(count)
            count = count + 1

        self.reacthts.append(Z0 + H)

        self.nsections = len(self.reacthts)
        self.nvolumes = self.nsections - 1
        self.meshz = nz * np.diff(self.reacthts)
        self.meshz = self.meshz.astype(int) + 1  # avoid zero mesh elements

        self.all_volumes = range(self.nvolumes)
        self.nonbaff_volumes = [
            sec for sec in self.all_volumes if sec not in self.baff_volumes
        ]
        self.nonstem_volumes = [
            0,
            1,
        ]  # this is 0,1 no matter how many impellers are there

        # note: stem_volumes include hub volumes also
        # these are volumes where we miss out polygon block
        self.stem_volumes = [
            sec for sec in self.all_volumes if sec not in self.nonstem_volumes
        ]

        # removes hub_volumes here for declaring patches
        self.only_stem_volumes = [
            sec for sec in self.stem_volumes if sec not in self.hub_volumes
        ]

        # to define mrf region
        # not that [1] is not a stem volume but baffles are there
        self.mrf_volumes = [1] + self.stem_volumes

        # increase grid points in the impeller section
        for i in self.baff_volumes:
            self.meshz[i] *= 2

        self.meshr = nr * np.diff(self.circradii)

        # adding polygon to hub mesh resolution
        self.meshr = np.append(nr * polyrad, self.meshr)
        self.meshr = self.meshr.astype(int)
        self.meshr += 1  # to avoid being zero

        self.centeroffset = 1  # one point on the axis
        self.polyoffset = self.nsplits  # number of points on polygon
        self.npts_per_section = (
            self.centeroffset + self.polyoffset + self.ncirc * self.nsplits
        )  # center+polygon+circles

    @classmethod
    def from_file(cls, yamlfile):
        if ".yaml" not in yamlfile:
            yamlfile += ".yaml"
        in_dict = parse_yaml(yamlfile)
        react_dict = {**in_dict["geometry"], **in_dict["mesh"]}
        return cls(**react_dict)



import json
import os
from pathlib import Path

from ruamel.yaml import YAML

from bird import logger


def _check_for_tabs_in_yaml(file_path: str) -> None:
    """
    Checks if a YAML file contains any tab characters.
    Raises a ValueError if tabs found.

    Parameters
    ----------
    file_path: str
        path to yaml file
    """

    with open(file_path, "r", encoding="utf-8") as f:
        lines = f.readlines()
    for iline, line in enumerate(lines):
        if "\t" in line:
            error_msg = (
                f"Tab character found on line {iline} of '{file_path}'. "
            )
            error_msg += "\nYAML files must use spaces for indentation."
            logger.error(error_msg)
            raise ValueError(error_msg)


def parse_json(file_path: str) -> dict:
    """
    Parse a json file into a dictionary

    Parameters
    ----------
    file_path: str
        path to json file

    Returns
    ----------
    inpt: dict
        Dictionary that contains the json file content
    """
    if not os.path.exists(file_path):
        raise FileNotFoundError(file_path)
    with open(file_path) as f:
        inpt = json.load(f)
    return inpt


def parse_yaml(file_path: str) -> dict:
    """
    Parse a yaml file into a dictionary

    Parameters
    ----------
    file_path: str
        path to yaml file

    Returns
    ----------
    inpt: dict
        Dictionary that contains the yaml file content
    """

    if not os.path.exists(file_path):
        raise FileNotFoundError(file_path)

    # Make sure the yaml file does not contain tabs
    _check_for_tabs_in_yaml(file_path)

    # Parse
    yaml = YAML(typ="safe")
    inpt = yaml.load(Path(file_path))
    return inpt



from typing import Callable

import numpy as np

from bird import logger


def conditional_average(
    x: np.ndarray, y: np.ndarray, nbins: int = 32
) -> tuple[np.ndarray, np.ndarray]:
    """
    Compute a 1D conditional average of y with respect to x
    The conditional average is distributed to neighbors of the binned array when needed

    Parameters
    ----------
    x: np.ndarray
        1D array with respect to which conditional averaged is performed
    y : np.ndarray
        1D array conditioned
    nbins: int
        Number of bins through x

    Returns
    ----------
    x_cond: np.ndarray
        The binned array of values conditioned againsts
    y_cond: np.ndarray
        The conditional averages at each bin
    """
    # Check the shape of input arrays
    try:
        assert len(x.shape) <= 2
        assert len(y.shape) <= 2
        if len(x.shape) == 2:
            assert x.shape[1] == 1
        if len(y.shape) == 2:
            assert y.shape[1] == 1
    except AssertionError:
        error_msg = "conditional average of tensors is ambiguous"
        error_msg += f"\nx shape =  {x.shape}"
        error_msg += f"\ny shape =  {y.shape}"
        logger.error(error_msg)
        raise AssertionError(error_msg)
    if len(x.shape) == 2:
        x = x[:, 0]
    if len(y.shape) == 2:
        y = y[:, 0]
    try:
        assert len(x) == len(y)
    except AssertionError:
        error_msg = "conditional average x and y have different dimension"
        error_msg += f"\ndim x =  {len(x)}"
        error_msg += f"\ndim y =  {len(y)}"
        logger.error(error_msg)
        raise AssertionError(error_msg)

    # Bin conditional space
    mag = np.amax(x) - np.amin(x)
    x_bin = np.linspace(
        np.amin(x) - mag / (2 * nbins), np.amax(x) + mag / (2 * nbins), nbins
    )
    weight = np.zeros(nbins)
    weightVal = np.zeros(nbins)
    asum = np.zeros(nbins)
    bsum = np.zeros(nbins)
    avalsum = np.zeros(nbins)
    bvalsum = np.zeros(nbins)
    inds = np.digitize(x, x_bin)

    a = abs(y - x_bin[inds - 1])
    b = abs(y - x_bin[inds])
    c = a + b
    a = a / c
    b = b / c

    # Conditional average at each bin
    for i in range(nbins):
        asum[i] = np.sum(a[np.argwhere(inds == i)])
        bsum[i] = np.sum(b[np.argwhere(inds == i + 1)])
        avalsum[i] = np.sum(
            a[np.argwhere(inds == i)] * y[np.argwhere(inds == i)]
        )
        bvalsum[i] = np.sum(
            b[np.argwhere(inds == i + 1)] * y[np.argwhere(inds == i + 1)]
        )
    weight = asum + bsum
    weightVal = avalsum + bvalsum

    # Assemble output
    x_cond = x_bin
    y_cond = weightVal / (weight)

    return x_cond, y_cond


def bissection(
    fun_val: float,
    nonlin_fun: Callable,
    num_iter: int = 1000,
    x_min: float = 1e-6,
    x_max: float = 1e6,
) -> float:
    """
    Solve a 1D non linear equation with a bissection method
    This is useful for adjusting the mesh grading so that mesh size varies smoothly

    Parameters
    ----------
    fun_val:
        Target value for the non linear function
    nonlin_fun:
        Non linear function
    num_iter: int
        Number of bissection iterations
        Defaults to 1000
    x_min : float
        Lower bound of the search interval
        Defaults to 1e-6
    x_max : float
        Upper bound of the search interval
        Defaults to 1e6


    Returns
    ----------
    x_mid: float
        The argument that achieves the desired function value
    """

    # Make sure residual sign changes over the search interval
    residual_min = nonlin_fun(x_min) - fun_val
    residual_max = nonlin_fun(x_max) - fun_val
    if residual_min * residual_max > 0:
        error_msg = "No guaranteed bissection solution"
        error_msg += (
            "\nSearch interval [{x_min:.4g}, {x_max:.4g}] may be too narrow"
        )
        raise ValueError(error_msg)

    # Do the bissection search
    for i in range(num_iter):
        x_mid = 0.5 * (x_max + x_min)
        residual_mid = nonlin_fun(x_mid) - fun_val
        if residual_mid * residual_max < 0:
            x_min = x_mid
            residual_min = residual_mid
        else:
            x_max = x_mid
            residual_max = residual_mid

    return x_mid



import argparse
import os
import pickle

import numpy as np

from bird.postprocess.post_quantities import *
from bird.utilities.ofio import *

parser = argparse.ArgumentParser(
    description="Compute means QoI of OpenFOAM fields"
)
parser.add_argument(
    "-f",
    "--caseFolder",
    type=str,
    metavar="",
    required=True,
    help="caseFolder to analyze",
    default=None,
)
parser.add_argument(
    "-vl",
    "--var_list",
    nargs="+",
    help="List of variables to compute",
    default=[
        "gh",
        "d",
        "CO2_liq",
        "CO_liq",
        "H2_liq",
        "c_CO2_liq",
        "c_CO_liq",
        "c_H2_liq",
    ],
    required=False,
)
parser.add_argument(
    "-avg",
    "--windowAve",
    type=int,
    metavar="",
    required=False,
    help="Window Average",
    default=1,
)
parser.add_argument(
    "-conv",
    "--windowConv",
    type=int,
    metavar="",
    required=False,
    help="Window Convergence",
    default=1,
)
parser.add_argument(
    "-dv",
    "--diff_val_list",
    nargs="+",
    type=float,
    help="List of diffusivities",
    default=[],
    required=False,
)
parser.add_argument(
    "-dn",
    "--diff_name_list",
    nargs="+",
    type=str,
    help="List of diffusivities",
    default=[],
    required=False,
)
args = parser.parse_args()


case_folder = args.caseFolder
var_name_list = args.var_list
time_float_sorted, time_str_sorted = get_case_times(case_folder)
cell_centers, _ = read_cell_centers(case_folder)
nCells = len(cell_centers)
diff_val_list = args.diff_val_list
diff_name_list = args.diff_name_list
assert len(diff_val_list) == len(diff_name_list)

window_ave = min(args.windowAve, len(time_str_sorted) - 1)
window_conv = min(args.windowConv, len(time_str_sorted) - 1)

variables = {}
variables_conv = {}
for name in var_name_list:
    variables_conv[name] = {}
    variables_conv[name]["x"] = []
    variables_conv[name]["y"] = []


def get_var(case_folder, time_folder, cell_centers, nCells, val_dict, name):
    if name.lower() == "gh":
        var, val_dict = compute_gas_holdup(
            case_folder,
            time_folder,
            nCells,
            field_dict=val_dict,
        )
    elif name.lower() == "d":
        var, val_dict = compute_ave_bubble_diam(
            case_folder,
            time_folder,
            nCells,
            field_dict=val_dict,
        )
    elif name.lower() == "co2_liq":
        var, val_dict = compute_ave_y_liq(
            case_folder,
            time_folder,
            nCells,
            species_name="CO2",
            field_dict=val_dict,
        )
    elif name.lower() == "co_liq":
        var, val_dict = compute_ave_y_liq(
            case_folder,
            time_folder,
            nCells,
            species_name="CO",
            field_dict=val_dict,
        )
    elif name.lower() == "h2_liq":
        var, val_dict = compute_ave_y_liq(
            case_folder,
            time_folder,
            nCells,
            species_name="H2",
            field_dict=val_dict,
        )
    elif name.lower() == "c_co2_liq":
        var, val_dict = compute_ave_conc_liq(
            case_folder,
            time_folder,
            nCells,
            species_name="CO2",
            field_dict=val_dict,
        )
    elif name.lower() == "c_co_liq":
        var, val_dict = compute_ave_conc_liq(
            case_folder,
            time_folder,
            nCells,
            species_name="CO",
            field_dict=val_dict,
        )
    elif name.lower() == "c_h2_liq":
        var, val_dict = compute_ave_conc_liq(
            case_folder,
            time_folder,
            nCells,
            species_name="H2",
            field_dict=val_dict,
        )
    else:
        raise NotImplementedError(f"Unknown variable {name}")

    return var, val_dict


print(f"Case : {case_folder}")

print(f"Window Ave")
for i_ave in range(window_ave):
    time_folder = time_str_sorted[-i_ave - 1]
    print(f"\tTime : {time_folder}")
    case_variables = []
    val_dict = {}
    for name in var_name_list:
        var, val_dict = get_var(
            case_folder,
            time_folder,
            cell_centers,
            nCells,
            val_dict=val_dict,
            name=name,
        )

        if i_ave == 0:
            variables[name] = var / window_ave
        else:
            variables[name] += var / window_ave


print(f"Window Conv")
for i_conv in range(window_conv):
    time_folder = time_str_sorted[-window_conv + i_conv]
    print(f"\tTime : {time_folder}")
    case_variables = []
    val_dict = {}
    for name in var_name_list:
        var, val_dict = get_var(
            case_folder,
            time_folder,
            cell_centers,
            nCells,
            val_dict=val_dict,
            name=name,
        )
        variables_conv[name]["x"] += [float(time_folder)]
        variables_conv[name]["y"] += [var]


with open(os.path.join(case_folder, "qoi.pkl"), "wb") as f:
    pickle.dump(variables, f)
with open(os.path.join(case_folder, "qoi_conv.pkl"), "wb") as f:
    pickle.dump(variables_conv, f)



import numpy as np
import stl
from scipy.spatial import Delaunay

from bird import logger


class STLMesh:
    def __init__(
        self,
        vertices=None,
        faces=None,
        area=None,
        planar=None,
        normal_dir=None,
    ):
        self.vertices = vertices
        self.faces = faces
        self.area = area
        self.normal_dir = normal_dir
        self.planar = planar
        self.status = {}
        self.update()

    def update_status(self):
        if self.vertices is None:
            self.status["vertices"] = False
        elif self.vertices.shape[1] == 3 and len(self.vertices.shape) == 2:
            self.status["vertices"] = True
        else:
            self.status["vertices"] = False
        if self.faces is None:
            self.status["faces"] = False
        elif self.vertices.shape[1] == 3 and len(self.faces.shape) == 2:
            self.status["faces"] = True
        else:
            self.status["faces"] = False
        if self.area is None:
            self.status["area"] = False
        elif isinstance(self.area, float):
            self.status["area"] = True
        else:
            self.status["area"] = False
        if self.normal_dir is None:
            self.status["normal_dir"] = False
        elif isinstance(self.normal_dir, int):
            self.status["normal_dir"] = True
        else:
            self.status["normal_dir"] = False
        if self.planar is None:
            self.status["planar"] = False
        elif isinstance(self.planar, bool):
            self.status["planar"] = True
        else:
            self.status["planar"] = False

    def update(self):
        self.update_status()
        if (
            self.status["vertices"]
            and not self.status["normal_dir"]
            and self.status["planar"]
        ):
            logger.debug("\t\tUpdating normal_dir to")
            for i in range(3):
                A = np.where(self.vertices[:, i] == self.vertices[0, i])[0]
                if len(A) == self.vertices.shape[0]:
                    self.normal_dir = i
                    self.status["normal_dir"] = True
                    logger.debug(f"\t\t\t{i}")
                    break
        if (
            self.status["vertices"]
            and not self.status["faces"]
            and self.status["planar"]
        ):
            logger.debug("\t\tUpdating faces")
            points = np.zeros((self.vertices.shape[0], 2))
            count = 0
            for i in range(3):
                if not i == self.normal_dir:
                    points[:, count] = self.vertices[:, i]
                    count += 1
            tri = Delaunay(points)
            self.faces = np.array(tri.simplices)
            self.status["faces"] = True

        if self.status["vertices"] and self.status["faces"]:
            logger.debug("\t\tUpdating area")
            self.calc_area()
            self.status["area"] = True

    def to_stl(self):
        stlObj = stl.mesh.Mesh(
            np.zeros(self.faces.shape[0], dtype=stl.mesh.Mesh.dtype)
        )
        for i, f in enumerate(self.faces):
            for j in range(3):
                stlObj.vectors[i][j] = self.vertices[f[j], :]
        return stlObj

    def from_stl(self, stlObj):
        self.vertices = np.unique(np.reshape(stlObj.vectors, (-1, 3)), axis=0)
        self.faces = np.zeros((stlObj.vectors.shape[0], 3), dtype=int)
        for i in range(stlObj.vectors.shape[0]):
            for j in range(stlObj.vectors.shape[1]):
                ind = np.argwhere(
                    np.linalg.norm(
                        self.vertices - stlObj.vectors[i, j], axis=1
                    )
                    == 0
                )[0][0]
                self.faces[i, j] = ind

    def calc_area(self):
        stlObj = self.to_stl()
        stlObj.update_areas()
        self.area = np.sum(stlObj.areas)

    def from_mesh_list(self, mesh_list):
        if len(mesh_list) > 1:
            self.planar = False
            self.update_status()

        offset = int(0)
        self.area = 0
        for patch_mesh in mesh_list:
            patch_mesh.faces += offset
            offset += len(patch_mesh.vertices)
            if self.vertices is None:
                self.vertices = patch_mesh.vertices
            else:
                self.vertices = np.vstack((self.vertices, patch_mesh.vertices))
            if self.faces is None:
                self.faces = patch_mesh.faces
            else:
                self.faces = np.vstack((self.faces, patch_mesh.faces))
            self.area += patch_mesh.area
        self.update()

    def rotate(self, theta=0, normal_dir=1):
        stlObj = self.to_stl()
        normal = [0, 0, 0]
        normal[normal_dir] = 1
        stlObj.rotate(normal, theta)
        return self.from_stl(stlObj)

    def translate(self, vector=np.array([0, 0, 0])):
        self.vertices += vector

    def save(self, filename):
        stlObj = self.to_stl()
        stlObj.save(f"{filename}", mode=stl.Mode.ASCII)



import random
import warnings

import numpy as np
import optuna
import pandas as pd
from scipy.interpolate import RBFInterpolator
from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import KFold, cross_val_score
from sklearn.preprocessing import OneHotEncoder
from tensorflow.keras.layers import Dense
from tensorflow.keras.models import Model as tfModel
from tensorflow.keras.models import Sequential

warnings.filterwarnings("ignore")

from bird import logger


def check_data_shape(X: np.ndarray, y: np.ndarray) -> None:
    """
    Tune the shape parameter (epsilon) of the multiquadratic RBF

    Parameters
    ----------
    X : np.ndarray
        Design configuration
        Dimension N by d, where N is the number of simulations,
                       and d is the number of design variables
    y: np.ndarray
        Array of quantity of interest
        Dimension N by 1, where N is the number of simulations
    """

    # Same number of samples
    assert X.shape[0] == y.shape[0]
    # Arrays are 2 dimensional
    assert len(X.shape) == 2
    assert len(y.shape) == 2
    # only 1 QoI
    assert y.shape[1] == 1
    logger.info(f"{X.shape[0]} sim with {X.shape[1]} design variables")


def tune_rbf(X: np.ndarray, y: np.ndarray) -> dict:
    """
    Tune the shape parameter (epsilon) of the multiquadratic RBF

    Parameters
    ----------
    X : np.ndarray
        Design configuration
        Dimension N by d, where N is the number of simulations,
                       and d is the number of design variables
    y: np.ndarray
        Array of quantity of interest
        Dimension N by 1, where N is the number of simulations
    Returns
    ----------
    params: dict
        dictionary of optimal parameters

    """

    check_data_shape(X=X, y=y)

    # Tune the RBFInterpolator with optuna
    # kernel - "multiquadric" (Can try other kernels)
    def objective(trial):
        epsilon = trial.suggest_float("epsilon", 0.1, 10.0, log=False)
        try:
            kf = KFold(n_splits=5, shuffle=True, random_state=42)
            return np.mean(
                [
                    mean_squared_error(
                        y[test],
                        RBFInterpolator(
                            X[train],
                            y[train],
                            epsilon=epsilon,
                            kernel="multiquadric",
                        )(X[test]),
                    )
                    for train, test in kf.split(X)
                ]
            )
        except:
            return float("inf")

    study = optuna.create_study(direction="minimize")
    study.optimize(objective, n_trials=20)
    return study.best_params


def tune_rf(X: np.ndarray, y: np.ndarray):
    """
    Tune the number of trees (n_estimators)
             tree depth (max_depth)
             number of samples in a leaf (min_samples_leaf)
    of the RandomForestRegressor

    Parameters
    ----------
    X : np.ndarray
        Design configuration
        Dimension N by d, where N is the number of simulations,
                       and d is the number of design variables
    y: np.ndarray
        Array of quantity of interest
        Dimension N by 1, where N is the number of simulations

    Returns
    ----------
    params: dict
        dictionary of optimal parameters

    """

    check_data_shape(X=X, y=y)

    # Tune the Random Forest
    def objective(trial):
        model = RandomForestRegressor(
            n_estimators=trial.suggest_int("n_estimators", 50, 200),
            max_depth=trial.suggest_int("max_depth", 3, 10),
            min_samples_leaf=trial.suggest_int("min_samples_leaf", 1, 10),
            random_state=42,
        )
        scores = cross_val_score(
            model, X, y, cv=5, scoring="neg_mean_squared_error"
        )
        return -np.mean(scores)

    study = optuna.create_study(direction="minimize")
    study.optimize(objective, n_trials=20)
    return study.best_params


def tune_nn(X_encoded: np.ndarray, y: np.ndarray) -> dict:
    """
    Tune the number of neurons (n_units)
             number of layers (n_layers)

    in a leaf (min_samples_leaf) of the RandomForestRegressor

    Parameters
    ----------
    X_encoded : np.ndarray
        Design configuration
        Dimension N by d, where N is the number of simulations,
                       and d is the number of design variables
    y: np.ndarray
        Array of quantity of interest
        Dimension N by 1, where N is the number of simulations

    Returns
    ----------
    params: dict
        dictionary of optimal parameters

    """
    check_data_shape(X=X_encoded, y=y)

    # Tune the Neural Network
    def objective(trial):
        units = trial.suggest_int("n_units", 16, 128)
        layers = trial.suggest_int("n_layers", 1, 3)
        kf = KFold(n_splits=5, shuffle=True, random_state=42)
        mses = []
        for train, test in kf.split(X_encoded):
            model = Sequential()
            model.add(
                Dense(
                    units, activation="relu", input_shape=(X_encoded.shape[1],)
                )
            )
            for _ in range(layers - 1):
                model.add(Dense(units, activation="relu"))
            model.add(Dense(1))
            model.compile(optimizer="adam", loss="mse")
            model.fit(
                X_encoded[train],
                y[train],
                epochs=100,
                batch_size=16,
                verbose=0,
            )
            mses.append(
                mean_squared_error(
                    y[test], model.predict(X_encoded[test]).flatten()
                )
            )
        return np.mean(mses)

    study = optuna.create_study(direction="minimize")
    study.optimize(objective, n_trials=20)
    return study.best_params


class Surrogate_wrapper:
    """
    Wrapper that builds the surrogate model and predicts the QOI value for given X
    """

    def __init__(
        self, model_type: str, X: np.ndarray, y: np.ndarray, params: dict
    ):
        """
        Create the surroagte wrapper

        Parameters
        ----------
        model_type : str
            'rbf' for RBFInterpolator
            'rf' for Random Forest
            'nn' for Neural Network

        X: np.ndarray
            Raw design configuration
            Dimension N by d, where N is the number of simulations,
                           and d is the number of design variables
        y: np.ndarray
            Raw array of quantity of interest
            Dimension N by 1, where N is the number of simulations
        params: dict
            dictionary of surrogate parameters

        """
        self.model_type = model_type
        self.encoder = None

        if model_type.lower() == "rbf":
            self.model = RBFInterpolator(
                X, y, kernel="multiquadric", epsilon=params["epsilon"]
            )
        elif model_type.lower() == "rf":
            self.model = RandomForestRegressor(**params, random_state=42)
            self.model.fit(X, y)
        elif model_type.lower() == "nn":
            self.encoder = OneHotEncoder(sparse_output=False)
            X_encoded = self.encoder.fit_transform(X)
            self.model = self._build_nn(
                X_encoded.shape[1], params["n_units"], params["n_layers"]
            )
            self.model.fit(X_encoded, y, epochs=100, batch_size=16, verbose=0)
        else:
            raise NotImplementedError

    def _build_nn(self, input_dim: int, units: int, layers: int) -> tfModel:
        """
        Builds the neural net

        Parameters
        ----------
        input_dim : int
            number of features
        units: int
            number of neurons per layer
        layers: int
            number of hidden layers

        Returns
        ----------
        model: tfModel
        """
        model = Sequential()
        model.add(Dense(units, activation="relu", input_shape=(input_dim,)))
        for _ in range(layers - 1):
            model.add(Dense(units, activation="relu"))
        model.add(Dense(1))
        model.compile(optimizer="adam", loss="mse")
        return model

    def predict(self, X):
        X = X.reshape(1, -1)
        if self.model_type == "nn":
            X = self.encoder.transform(X)
            return float(self.model.predict(X)[0])
        elif self.model_type == "rf":
            return float(self.model.predict(X)[0])
        else:
            return float(self.model(X)[0])



import corner
import jax
from jax._src import config

config.update("jax_platforms", "cpu")

import jax.numpy as jnp
import jax.random as random
import numpy as np
from packaging import version

if version.parse(jax.__version__) >= version.parse("0.7.0"):
    # https://github.com/pyro-ppl/numpyro/issues/2051
    import jax.experimental.pjit
    from jax.extend.core.primitives import jit_p

    jax.experimental.pjit.pjit_p = jit_p

import numpyro
import numpyro.distributions as dist
from numpyro.infer import MCMC, NUTS
from prettyPlot.plotting import *

from bird import logger


def c_model(
    theta: list, t: np.ndarray | jnp.ndarray, t0: float, c0: float
) -> np.ndarray | jnp.ndarray:
    """
    Model of concentration vs time function involving kLa
    """
    cstar, kla = theta
    return (cstar - c0) * (1.0 - jnp.exp(-kla * (t - t0))) + c0


def bayes_step(
    time_obs: np.ndarray | jnp.ndarray,
    concentration_obs: np.ndarray | jnp.ndarray,
):
    """
    Prior and likelihood definition for MCMC sampling
    """
    # define parameters (incl. prior ranges)
    cstar = numpyro.sample("cstar", dist.Uniform(1e-8, 1000))
    kla = numpyro.sample("kla", dist.Uniform(1e-8, 1000))
    sigma = numpyro.sample("sigma", dist.Uniform(1e-8, 1))

    # initial values
    t0 = time_obs[0]
    c0 = concentration_obs[0]

    # concentration predictions
    c_pred = c_model([cstar, kla], t=time_obs, t0=t0, c0=c0)

    # likelihood uncertainty (assumed homoskedastic)
    std_c = jnp.ones(c_pred.shape[0]) * sigma

    # MCMC sampling with multivariate normal
    numpyro.sample("obs", dist.Normal(c_pred, std_c), obs=concentration_obs)


def hmc_samples_2_np_samples(hmc_samples: dict) -> np.ndarray:
    """
    Convert hmc samples generated by numpyro to numpy
    """
    labels = list(hmc_samples.keys())
    nsamples = len(hmc_samples[labels[0]])
    nparams = len(labels)
    np_hmc_samples = np.zeros((nsamples, nparams))
    labels_np = ["cstar", "kla", "sigma"]

    for ilabel, label in enumerate(labels):
        if label == "cstar":
            nplabel = labels_np.index("cstar")
        if label == "kla":
            nplabel = labels_np.index("kla")
        if label == "sigma":
            nplabel = labels_np.index("sigma")
        np_hmc_samples[:, nplabel] = np.array(hmc_samples[label])

    return np_hmc_samples


def np_samples_2_pred(
    np_hmc_samples: np.ndarray,
    time_obs: np.ndarray,
    concentration_obs: np.ndarray,
) -> dict:
    """
    Go from parameter samples to predictions.
    Useful for accuracy calculation and plotting.
    """
    # Uncertainty propagation
    nsamples = np_hmc_samples.shape[0]
    realization = []
    for i in range(nsamples):
        c_pred = c_model(
            np_hmc_samples[i, :-1], time_obs, time_obs[0], concentration_obs[0]
        )
        realization.append(c_pred)
    realization = np.array(realization)

    mean_real = np.mean(realization, axis=0)
    p90_real = np.percentile(realization, 90, axis=0)
    p10_real = np.percentile(realization, 10, axis=0)

    return {"mean_pred": mean_real, "p10_pred": p10_real, "p90_pred": p90_real}


def read_kla_cstar(hmc_samples: dict) -> dict:
    """
    go from hmc samples generated by numpyro to relevant arrays
    """
    np_hmc_samples = hmc_samples_2_np_samples(hmc_samples)
    return {
        "kla_samples": np_hmc_samples[:, 1],
        "cstar_samples": np_hmc_samples[:, 0],
    }


def compute_accuracy(
    hmc_samples: dict, time_obs: np.ndarray, concentration_obs: np.ndarray
) -> float:
    """
    Compute how well the sampled parameter explain the observed concentration data
    """
    np_hmc_samples = hmc_samples_2_np_samples(hmc_samples)
    pred_dict = np_samples_2_pred(np_hmc_samples, time_obs, concentration_obs)
    mean_real = pred_dict["mean_pred"]

    return np.mean(
        abs(mean_real - concentration_obs)
        / np.clip(concentration_obs, a_min=1e-12, a_max=None)
    )


def check_data_size(time_obs: np.ndarray, concentration_obs: np.ndarray):
    """
    Sanity checks
    """
    assert time_obs.shape == concentration_obs.shape
    assert len(time_obs) > 5
    assert len(time_obs.shape) == 1


def compute_kla(
    data_t: np.ndarray,
    data_c: np.ndarray,
    num_warmup: int = 4000,
    num_samples: int = 1000,
    max_chop: int | None = None,
    bootstrap: bool = True,
) -> dict:
    """
    Main loop to compute kla
    """
    rng_key = random.PRNGKey(0)
    rng_key, rng_key_ = random.split(rng_key)

    acc = []
    kla = []
    cstar = []
    kla_err = []
    cstar_err = []
    ind = []
    check_data_size(data_t, data_c)

    data_t_tmp = data_t.copy()
    data_c_tmp = data_c.copy()

    # Find where to start in the timeseries
    for ind_start in range(len(data_t_tmp) - 5):
        if max_chop is not None:
            max_chop -= 1
        logger.debug(f"Chopping index = {ind_start}")
        data_t = data_t_tmp[ind_start:]
        data_c = data_c_tmp[ind_start:]

        # Hamiltonian Monte Carlo (HMC) with no u turn sampling (NUTS)
        kernel = NUTS(bayes_step, target_accept_prob=0.9)
        mcmc = MCMC(
            kernel,
            num_warmup=num_warmup,
            num_samples=num_samples,
            progress_bar=False,
        )
        mcmc.run(
            rng_key_,
            time_obs=jnp.array(data_t),
            concentration_obs=jnp.array(data_c),
        )
        # mcmc.print_summary()

        # Read samples
        hmc_samples = mcmc.get_samples()
        samp_dict = read_kla_cstar(hmc_samples)
        kla_samples = samp_dict["kla_samples"]
        cstar_samples = samp_dict["cstar_samples"]

        accuracy = compute_accuracy(hmc_samples, data_t, data_c)

        ind.append(ind_start)
        acc.append(accuracy)
        kla.append(np.mean(kla_samples))
        cstar.append(np.mean(cstar_samples))
        kla_err.append(np.std(kla_samples))
        cstar_err.append(np.std(cstar_samples))

        # post_cal(hmc_samples, data_t, data_c)

        # If accuracy has significantly improved, we have chopped enough
        if len(acc) >= 3 and abs(acc[-1] - acc[-2]) < 0.1 * acc[0]:
            break
        # If we exceed the maximum number of timesteps to ignore, break
        if max_chop == 0:
            break

    bootstrapped = False
    # Data bootstrapping
    ## Include the uncertainty due to the number of data point available
    ## Reduce the number of data points following 4 arbitrary scenarios
    ## scenario 1: remove the first point
    ## scenario 2: remove the first 2 points
    ## scenario 3: remove the last point
    ## scenario 4: remove the last 2 points

    # Don't bootstrap if there aren't enough data points
    if bootstrap and len(data_t_tmp[ind[-1] :]) > 10:
        kla_boot = []
        kla_err_boot = []
        cstar_boot = []
        cstar_err_boot = []
        bootstrapped = True
        logger.info("Doing data bootstrapping")

        for i in range(4):
            logger.info(f"\t scenario {i}")
            if i == 0:
                data_t = data_t_tmp[ind_start + 1 :]
                data_c = data_c_tmp[ind_start + 1 :]
            elif i == 1:
                data_t = data_t_tmp[ind_start + 2 :]
                data_c = data_c_tmp[ind_start + 2 :]
            elif i == 2:
                data_t = data_t_tmp[ind_start:-1]
                data_c = data_c_tmp[ind_start:-1]
            elif i == 3:
                data_t = data_t_tmp[ind_start:-2]
                data_c = data_c_tmp[ind_start:-2]

            # Hamiltonian Monte Carlo (HMC) with no u turn sampling (NUTS)
            kernel = NUTS(bayes_step, target_accept_prob=0.9)
            mcmc = MCMC(
                kernel,
                num_warmup=num_warmup,
                num_samples=num_samples,
                progress_bar=False,
            )
            mcmc.run(
                rng_key_,
                time_obs=jnp.array(data_t),
                concentration_obs=jnp.array(data_c),
            )
            # mcmc.print_summary()

            # Read samples
            hmc_samples = mcmc.get_samples()
            samp_dict = read_kla_cstar(hmc_samples)
            kla_samples = samp_dict["kla_samples"]
            cstar_samples = samp_dict["cstar_samples"]

        kla_boot.append(np.mean(kla_samples))
        kla_err_boot.append(np.std(kla_samples))
        cstar_boot.append(np.mean(cstar_samples))
        cstar_err_boot.append(np.std(cstar_samples))

    else:
        kla_boot = [kla[-1]]
        kla_err_boot = [kla_err[-1]]
        cstar_boot = [cstar[-1]]
        cstar_err_boot = [cstar_err[-1]]

    return {
        "kla": np.mean(np.array(kla_boot)),
        "kla_err": np.mean(np.array(kla_err_boot)),
        "cstar": np.mean(np.array(cstar_boot)),
        "cstar_err": np.mean(np.array(cstar_err_boot)),
        "kla_noboot": np.mean(np.array([kla[-1]])),
        "kla_err_noboot": np.mean(np.array([kla_err[-1]])),
        "cstar_noboot": np.mean(np.array([cstar[-1]])),
        "cstar_err_noboot": np.mean(np.array([cstar_err[-1]])),
        "bootstrapped": bootstrapped,
    }


def post_cal(
    hmc_samples: dict, time_obs: np.ndarray, concentration_obs: np.ndarray
) -> None:
    """
    Plot MCMC results
    """
    np_hmc_samples = hmc_samples_2_np_samples(hmc_samples)
    labels_np = ["cstar", "kla", "sigma"]
    nsamples = np_hmc_samples.shape[0]
    labels = list(hmc_samples.keys())
    nparams = len(labels)

    # Distribution of the parameter samples
    fig = corner.corner(np_hmc_samples, labels=labels_np)

    # Convergence
    fig, axes = plt.subplots(nparams, sharex=True)
    for i in range(nparams):
        ax = axes[i]
        ax.plot(np_hmc_samples[:, i], "k", alpha=0.3, rasterized=True)
        ax.set_ylabel(labels[i])

    # Uncertainty propagation
    pred_dict = np_samples_2_pred(np_hmc_samples, time_obs, concentration_obs)
    mean_real = pred_dict["mean_pred"]
    p10_real = pred_dict["p10_pred"]
    p90_real = pred_dict["p90_pred"]

    fig = plt.figure()
    plt.plot(
        time_obs, mean_real, color="k", linewidth=3, label="mean predictions"
    )
    plt.plot(
        time_obs,
        p90_real,
        "--",
        color="k",
        linewidth=3,
        label="10th and 90th percentile",
    )
    plt.plot(time_obs, p10_real, "--", color="k", linewidth=3)
    plt.plot(
        time_obs,
        concentration_obs,
        "o",
        color="k",
        markersize=7,
        label="Data observed",
    )
    pretty_labels("t", "c[mol/m3]", 14)
    pretty_legend()
    plt.show()


def print_res_dict(res_dict: dict) -> None:
    """
    Log kla and cstar to screen
    """
    kla = res_dict["kla"]
    kla_err = res_dict["kla_err"]
    cs = res_dict["cstar"]
    cs_err = res_dict["cstar_err"]

    kla_nb = res_dict["kla_noboot"]
    kla_err_nb = res_dict["kla_err_noboot"]
    cs_nb = res_dict["cstar_noboot"]
    cs_err_nb = res_dict["cstar_err_noboot"]

    bs = res_dict["bootstrapped"]

    if bs:
        logger.info(f"\tkla = {kla*3600:.4g} +/- {kla_err*3600:.4g} [h-1]")
        logger.info(f"\tcstar = {cs:.4g} +/- {cs_err:.4g} [mol/m3]")
        logger.info(f"Without data bootstrap")
    logger.info(f"\tkla = {kla_nb*3600:.4g} +/- {kla_err_nb*3600:.4g} [h-1]")
    logger.info(f"\tcstar = {cs_nb:.4g} +/- {cs_err_nb:.4g} [mol/m3]")

